tempo = new TempoTool(logger: logger, componentAccessor: componentAccessor, issues: issues, user: user, apiClientFactory: apiClientFactory)

/**
 * Tool to get Tempo data from the Tempo Cloud REST API.
 *
 * @see https://tempo-io.github.io/tempo-api-docs/
 */
class TempoTool {
	def logger

	/* Tempo REST API authentication. */
	def tempoAccessToken = '' // Tempo API access token, see: https://www.midori-global.com/products/better-excel-exporter-for-jira/cloud/documentation/integrations/tempo-timesheets

	/* Configuration. */
	def dateFrom = '2000-01-01' // start date for the exportable Tempo worklog period
	def dateTo = '2099-12-31' // end date for the exportable Tempo worklog period
	def tempoFilterByUser = false // set to "true" to export the Tempo worklogs of the current user only, or "false" to export everyone's

	/* Maximum number of issue keys to pass in the URL in a single REST API call. */
	def ISSUE_KEYS_PER_REQUEST = 50

	def apiClientFactory
	def componentAccessor
	def tempoClient

	def issues
	def user

	def preloadedAccounts = []
	def preloadedWorklogs = []
	def preloadedWorkAttributes = [:]

	def slurper = new groovy.json.JsonSlurper()

	/**
	 * Returns the account for the passed worklog (if it has work attribute of that type) or for the passed issue otherwise.
	 */
	def getAccount(issue, worklog) {
		def account

		// get value from "Account" type work attribute
		def accountWorkAttribute = getAllWorkAttributes().values().find { it.type == 'ACCOUNT' }
		if (accountWorkAttribute) {
			def accountKey = worklog.attributes.values?.find { it.key == accountWorkAttribute.key }?.value
			account = accountKey ? getAccountByKey(accountKey).name : null
		}

		// get value from "Account" type custom field
		if (!account) {
			def accountCustomField = componentAccessor.customFieldManager.getCustomFieldObjects(issue).find { it.customFieldType.key == 'com.atlassian.plugins.atlassian-connect-plugin:io.tempo.jira__account' }
			account = accountCustomField ? issue.getCustomFieldValue(accountCustomField)?.value : null
		}

		account
	}

	/**
	 * Returns the team for the passed issue.
	 */
	def getTeam(issue) {
		def teamCustomField = componentAccessor.customFieldManager.getCustomFieldObjects(issue).find { it.customFieldType.key == 'com.atlassian.plugins.atlassian-connect-plugin:io.tempo.jira__team' }
		teamCustomField ? issue.getCustomFieldValue(teamCustomField)?.value : null
	}

	/**
	 * Returns the worklogs for the passed issue as a <code>List</code> of <code>Maps</code>, or the exception object if something went wrong.
	 */
	def getWorklogs(issueKey) {
		if (!preloadedWorklogs) {
			issues.collect { it.key }.collate(ISSUE_KEYS_PER_REQUEST).each { issueKeys ->
				if (!(preloadedWorklogs instanceof Throwable)) {
					logger.log("Downloading Tempo worklogs for ${issueKeys.join(', ')}")
					def result = callTempoApi("/worklogs?from=${dateFrom}&to=${dateTo}&issue=${issueKeys.join('&issue=')}", 'results')
					(!(result instanceof Throwable)) ? preloadedWorklogs.addAll(result.results) : (preloadedWorklogs = result)
				}
			}
		}
		!(preloadedWorklogs instanceof Throwable) ?
				preloadedWorklogs.findAll { (it.issue.key == issueKey) && (!tempoFilterByUser || (user.accountId == it.author.accountId)) } :
				[[description: "Failed to download Tempo worklogs! (see Tempo API guide: https://www.midori-global.com/products/better-excel-exporter-for-jira/cloud/documentation/integrations/tempo-timesheets): ${preloadedWorklogs.message}"]]
	}

	/**
	 * Returns comma-separated work attribute name-value pairs for the passed worklog.
	 */
	def getWorkAttributes(worklog) {
		def workAttributes = getAllWorkAttributes()
		worklog.attributes.values?.findAll { it.value }.collect {
			def attributeName = workAttributes ? workAttributes[it.key].name : it.key
			def attributeValue = it.value
			if (workAttributes && workAttributes[it.key].type == 'ACCOUNT') {
				attributeValue = getAccountByKey(it.value)?.name ?: it.value
			}
			"${attributeName}: ${attributeValue}"
		}.join(", ")
	}

	private def getAccountByKey(accountKey) {
		getAllAccounts().find { it.key == accountKey }
	}

	/**
	 * Returns all accounts configured in Tempo.
	 */
	private def getAllAccounts() {
		if (!preloadedAccounts) {
			def result = callTempoApi('/accounts')
			preloadedAccounts = !(result instanceof Throwable) ? result.results : result
		}
		!(preloadedAccounts instanceof Throwable) ? preloadedAccounts : null
	}

	/**
	 * Returns all work attributes configured in Tempo.
	 */
	private def getAllWorkAttributes() {
		if (!preloadedWorkAttributes) {
			def result = callTempoApi('/work-attributes')
			preloadedWorkAttributes = !(result instanceof Throwable) ? result.results.collectEntries {
				[(it.key): it]
			} : result
		}
		!(preloadedWorkAttributes instanceof Throwable) ? preloadedWorkAttributes : null
	}

	/**
	 * Calls the Tempo Cloud REST API and handles paging if needed.
	 * If the entityListFieldName parameter is passed, then that property will contain all the data of the same property of each page in the final result.
	 */
	private def callTempoApi(url, entityListFieldName = null) {
		try {
			logger.log("Tempo URL: ${url}")
			if (!tempoAccessToken) {
				throw new IllegalStateException('Enable Tempo!')
			}
			if (!tempoClient) {
				tempoClient = apiClientFactory.createTempoClient(tempoAccessToken)
			}
			def jsonPages = tempoClient.getJson(url)
			// logger.log("Tempo JSON for ${url}: ${jsonPages}")

			def result
			jsonPages.collect {
				def pageData = slurper.parseText(it)
				if (!result) {
					result = pageData
				} else {
					result[entityListFieldName].addAll(pageData[entityListFieldName])
				}
			}

			return result
		} catch (Throwable e) {
			logger.log('Failed to download Tempo data', e)
			return e
		}
	}
}