zephyr = new ZephyrTool(logger: logger, apiClientFactory: apiClientFactory, userManager: userManager)

/**
 * Tool to get Zephyr test data from the Zephyr Cloud REST API.
 *
 * @see https://support.smartbear.com/zephyr-squad-cloud/docs/api/index.html
 */
class ZephyrTool {
	def logger

	/* Zephyr REST API authentication. */
	def zephyrAccessKey = '' // Zephyr API access key, see: https://www.midori-global.com/products/better-excel-exporter-for-jira/cloud/documentation/integrations/zephyr
	def zephyrSecretKey = ''
	def zephyrAccountId = ''

	def apiClientFactory
	def userManager
	def zephyrClient

	def slurper = new groovy.json.JsonSlurper()

	/**
	 * Returns the Zephyr test steps for the passed issue (as Groovy objects slurped from <code>https://zfjcloud.docs.apiary.io/#reference/teststep/create-and-get-all-teststep/get-all-teststeps-v1<code>)
	 * or the exception object if something went wrong.
	 */
	def getTestSteps(def issue) {
		if (!isTest(issue)) {
			return []
		}

		def issueId = issue.id
		def issueKey = issue.key
		def projectId = issue.projectObject.id
		logger.log("Downloading Zephyr test steps for issue #${issueId} [${issueKey}]")
		def testSteps = callZephyrApi("/public/rest/api/2.0/teststep/${issueId}?projectId=${projectId}", "testSteps")
		if (!(testSteps instanceof Throwable)) {
			logger.log("${testSteps.size()} Zephyr test steps found for issue #${issueId} [${issueKey}]")

			testSteps.testSteps?.each { testStep ->
				def attachmentNames = testStep.attachments?.collect { attachment -> attachment.name.trim() }
				testStep.attachmentCount = attachmentNames.size()
				testStep.attachmentsString = attachmentNames.join(', ')
			}

			return testSteps.testSteps
		}

		return [["orderId": "Failed to download Zephyr test steps! ${seeZephyrGuide()}: ${testSteps.message}"]]
	}

	/**
	 * Returns the Zephyr test executions for the passed issue (as Groovy objects slurped from <code>https://zfjcloud.docs.apiary.io/#reference/execution/get-list-of-executions/get-list-of-executions<code>)
	 * or the exception object if something went wrong.
	 */
	def getTestExecutions(def issue) {
		if (!isTest(issue)) {
			return []
		}

		def issueId = issue.id
		def issueKey = issue.key
		def projectId = issue.projectObject.id
		logger.log("Downloading Zephyr test executions for issue #${issueId} [${issueKey}]")
		def testExecutions = callZephyrApi("/public/rest/api/1.0/executions?projectId=${projectId}&issueId=${issueId}", "executions")
		if (!(testExecutions instanceof Throwable)) {
			logger.log("${testExecutions.executions.size()} Zephyr test executions found for issue #${issueId} [${issueKey}]")

			// resolve account IDs to users
			def accountIds = (testExecutions.executions.collect { it.execution.executedByAccountId } + testExecutions.executions.collect { it.execution.assignedToAccountId }).findAll { it != null }.unique()
			def users = accountIds ? userManager.getUsersByAccountId(accountIds) : []

			testExecutions.executions.each { execution ->
				def executedByUser = users.find { user -> user.accountId == execution.execution.executedByAccountId }
				execution.execution.executedBy = executedByUser?.displayName

				def assignedToUser = users.find { user -> user.accountId == execution.execution.assignedToAccountId }
				execution.execution.assignedTo = assignedToUser?.displayName

				execution.execution.executedOn = execution.execution.executedOn ? new Date((long) execution.execution.executedOn) : null
				execution.execution.defectsAsString = execution.execution.defects.collect { it.key }.join(", ")
				execution.execution.stepDefectsAsString = execution.execution.stepDefects.collect { it.key }.join(", ")
			}

			// sort
			testExecutions.executions = testExecutions.executions.sort { a, b -> a.execution.executedOn <=> b.execution.executedOn }.unique { a, b -> a.execution.id <=> b.execution.id } // FIXME drop this when Zephyr fixes their API (currently Zephyr returns an unordered list of executions with buggy pagination, therefore we need to sort and remove duplicates)
			return testExecutions.executions
		}

		return [["execution": ["cycleName": "Failed to download Zephyr test executions! ${seeZephyrGuide()}: ${testExecutions.message}"]]]
	}

	/**
	 * Returns the Zephyr test step results for the passed execution (as Groovy objects slurped from <code>https://zfjcloud.docs.apiary.io/#reference/stepresult/get-stepresults-by-execution/get-stepresults-by-execution<code>)
	 * or the exception object if something went wrong.
	 *
	 * Note that different executions of the same test may return different number of step results.
	 * It occurs if new steps have been added to or removed from the test between the executions.
	 * (The Zephyr UI always shows the list of the current steps, but the Zephyr REST API works differently.)
	 *
	 * @param expandTestSteps whether to include test step instructions, data, expected result, etc. in the return value.
	 */
	def getTestStepResults(def executionId, def issue, def expandTestSteps = false) {
		if (!isTest(issue)) {
			return []
		}

		def issueId = issue.id
		logger.log("Downloading Zephyr test step results for execution #${executionId}")
		def testStepResults = callZephyrApi("/public/rest/api/1.0/stepresult/search?executionId=${executionId}&issueId=${issueId}&isOrdered=true", "stepResults")
		if (!(testStepResults instanceof Throwable)) {
			logger.log("${testStepResults.size()} Zephyr test step results found for execution #${executionId}")
			def testSteps = (expandTestSteps && testStepResults.stepResults) ? getTestSteps(issue) : []
			testStepResults.stepResults?.each { stepResult ->
				stepResult.executedOn = stepResult.executedOn ? new Date((long) stepResult.executedOn) : null
				stepResult.defectsAsString = stepResult.defects.collect { it.key }.join(", ")
				if (expandTestSteps) {
					stepResult.step = testSteps.find { it.id.toString() == stepResult.stepId }
				}
			}
			return testStepResults.stepResults ?: []
		}

		return [["stepResults": ["cycleName": "Failed to download Zephyr test step results! ${seeZephyrGuide()}: ${testStepResults.message}"]]]
	}

	private static def isTest(def issue) {
		issue.issueTypeObject.name == 'Test'
	}

	private static def seeZephyrGuide() {
		'(see Zephyr API guide: https://www.midori-global.com/products/better-excel-exporter-for-jira/cloud/documentation/integrations/zephyr)'
	}

	/**
	 * Calls the Zephyr Cloud REST API and handles paging if needed.
	 * If the entityListFieldName parameter is passed, then that property will contain all the data of the same property of each page in the final result.
	 */
	private def callZephyrApi(url, entityListFieldName = null) {
		try {
			logger.log("Zephyr URL: ${url}")
			if (!zephyrAccessKey) {
				throw new IllegalStateException('Zephyr API access key is not set')
			}
			if (!zephyrClient) {
				zephyrClient = apiClientFactory.createZephyrClient(zephyrAccessKey, zephyrSecretKey, zephyrAccountId)
			}
			def jsonPages = zephyrClient.getJson(url)
			// logger.log("Zephyr JSON for ${url}: ${jsonPages}")

			def result
			jsonPages.collect {
				def pageData = slurper.parseText(it)
				if (!result) {
					result = pageData
				} else {
					result[entityListFieldName].addAll(pageData[entityListFieldName])
				}
			}

			return result
		} catch (Throwable e) {
			logger.log('Failed to download Zephyr test data', e)
			return e
		}
	}
}
